#include "mex.h"
#include "math.h"


/* A = B' */
void transposeMatrix(double *A,double *B,int m,int n){
    
    int c,r;
    int inds,indd; 
    
    for(c=0;c<n;c++){
        for(r=0;r<m;r++){			
            inds = c*m+r;
            indd = r*n+c;
            A[indd]=B[inds];
        }		
    }	
}

/*----------------------------------------------------------------------*/
/*                    CONSTANT LENGTH CONSTRAINTS                       */
/*----------------------------------------------------------------------*/

/*
 * leng := proc(x0,y0,z0,x1,y1,z1)
 *   local val: 
 *   val:=((x0-x1)^2+(y0-y1)^2+(z0-z1)^2-Length*Length):
 * end:
 * f:=leng(x0,y0,z0,x1,y1,z1):
 * g:=vector([diff(f,x0),diff(f,y0),diff(f,z0),diff(f,x1),diff(f,y1),diff(f,z1),f]):
 * C(g,'optimized');
 */
static void DLength2(double *f,double *g,double x0,double y0,double z0,double x1,double y1,double z1,double Length)
{
    double
    t1 = -x1+x0,
    t2 = -y1+y0,
    t3 = -z1+z0,
    t4 = t1*t1,
    t5 = t2*t2,
    t6 = t3*t3,
    t7 = Length*Length;
    if(f)
        *f=(t4+t5+t6-t7);
    if(g){
        g[0] = 2*t1;
        g[1] = 2*t2;
        g[2] = 2*t3;
        g[3] = -2*t1;
        g[4] = -2*t2;
        g[5] = -2*t3;
    }
    
    if(0){
        mexPrintf("point 1 = %f %f %f - point 2 = %f %f %f .. length = %f\n", x0, y0, z0, x1, y1, z1, f);
    }
}

/*
 * leng := proc(x0,y0,z0,x1,y1,z1)
 *   local val: 
 *   val:=sqrt((x1-x0)^2+(y1-y0)^2+(z1-z0)^2)-Length:
 *  end:
 */

/*DLength1(&Fl,Gl,Xs[i0],Ys[i0],Zs[i0],Xs[i1],Ys[i1],Zs[i1],ELength)*/;
static void DLength1(double *f,double *g,double x0,double y0,double z0,double x1,double y1,double z1,double Length)
{
    double 
    t1  = x1*x1,
    t4  = x0*x0,
    t5  = y1*y1,
    t8  = y0*y0,
    t9  = z1*z1,
    t12 = z0*z0,
    t14 = sqrt(t1-2.0*x1*x0+t4+t5-2.0*y1*y0+t8+t9-2.0*z1*z0+t12),
    t15 = 1/t14,
    t16 = -x1+x0,
    t19 = -y1+y0,
    t22 = -z1+z0;
    
    if(0){
        mexPrintf("point 1 = %f %f %f - point 2 = %f %f %f .. length = %f\n", x0, y0, z0, x1, y1, z1, t14-Length);
    }
            
    if(f)
        *f=t14-Length;
    if(g){
        g[0] = t15*t16;
        g[1] = t15*t19;
        g[2] = t15*t22;
        g[3] = -t15*t16;
        g[4] = -t15*t19;
        g[5] = -t15*t22;
    }
}

/*LengthConstraints(C,A,Coords,NVerts,Edges,NEdges,Lengths);*/
void LengthConstraints(double *C,double *A,double *Coords,int NVerts,double *Edges,int NEdges,double *Lengths,int squareP)
{
    double Fl,Gl[6],ELength=0.0;
    int i0,i1,j0,j1,k0,k1,EdgeI,DimA=3*NVerts,temp,a,b;
    
    if(A){
        for(temp=0;temp<DimA*NEdges;temp++){
            A[temp]=0;
        }    
    }
                 
    for (EdgeI=0;EdgeI<NEdges;EdgeI++){
        
        /* Transform matlab indices into C indices*/
        a = (int)Edges[EdgeI]-1;
        b = (int)Edges[EdgeI+NEdges]-1;
        
        
        i0=a*3;    
        i1=b*3;        
        j0=i0+1;
        j1=i1+1;
        k0=j0+1;
        k1=j1+1;
        
        if(0){
            mexPrintf("i0=%d j0=%d k0=%d i1=%d j1=%d k1=%d",i0,j0,k0,i1,j1,k1);
        }
        
        if(Lengths)
            ELength=Lengths[EdgeI];
        
        if(squareP>0){
            DLength2(&Fl,Gl,Coords[i0],Coords[j0],Coords[k0],Coords[i1],Coords[j1],Coords[k1],ELength);
        }
        else{
            DLength1(&Fl,Gl,Coords[i0],Coords[j0],Coords[k0],Coords[i1],Coords[j1],Coords[k1],ELength);            
        }
                
        if(0){
            mexPrintf("L=%f L0=%f \n",Fl,ELength);
        }
        
                
        if(A){
            A[i0]=Gl[0];
            A[j0]=Gl[1];
            A[k0]=Gl[2];
            A[i1]=Gl[3];
            A[j1]=Gl[4];
            A[k1]=Gl[5];
            A+=DimA;
        }
        if(C){
            C[0]=Fl;
            C++;
        }
    }
}

void printUsage(){

    mexPrintf("\n\nUSAGE:\n");
    mexPrintf("[F G] = LenghtConstraintsMex(XYZ,Edges,refLengths,squareP) \n");   
    mexPrintf("\n\nINPUTS:\n");
    mexPrintf("XYZ: 1xD matrix where D is the output dimensionality (3 x number of vertices) \n");
    mexPrintf("Edges: Mx2 matrix where M is the number of edges\n");
    mexPrintf("refLengths: Mx1 vector of reference edge lengths\n");
    mexPrintf("squareP: 1 if squared length equation is used, 0 otherwise\n");
    mexPrintf("\n\nOUTPUTS:\n");
    mexPrintf("F: Mx1 vector of edge length constraint violations\n");
    mexPrintf("G: MxD matrix of jacobian of the violation vector\n");
    mexPrintf("\n\n\n");
}


/*----------------------------------------------------------------------*/
/*                            MEX FUNCTION                              */
/*----------------------------------------------------------------------*/

void mexFunction( int nlhs, mxArray *plhs[],int nrhs, const mxArray *prhs[])
{ 
    double *C,*G;
    double *Coords,*Edges,*Lengths;
    int NVerts,NEdges,n2,n3;
    mxArray* GArray;
    
    /* Check for proper number of arguments */
    if (nrhs != 4) { 
        printUsage();
        mexErrMsgTxt("Four input arguments required."); 

    } else if (nlhs > 2) {
        printUsage();
        mexErrMsgTxt("Too many output arguments."); 
    } 
    
    Coords  = mxGetPr(prhs[0]);
    Edges   = mxGetPr(prhs[1]);
    Lengths = mxGetPr(prhs[2]);
    
   
    /* Sanity checks */
    n3      = mxGetNumberOfElements(prhs[0]);
    NVerts  = n3/3;
    n2      = mxGetNumberOfElements(prhs[1]);
    NEdges  = n2/2;

    if(Lengths&&(mxGetNumberOfElements(prhs[2])!=NEdges))
        mexErrMsgTxt("The array of edge lengths is not of the proper size");
    if(((3*NVerts)!=n3)||((2*NEdges)!=n2))
        mexErrMsgTxt("Coords or Edges arrays do not adequate dimensions.");
    
    /* Create a matrix for the return argument */ 
    plhs[0] = mxCreateDoubleMatrix(NEdges,1,mxREAL);    
    plhs[1] = mxCreateDoubleMatrix(NEdges,n3,mxREAL);
    
    GArray = mxCreateDoubleMatrix(n3,NEdges,mxREAL);
    
    C       = mxGetPr(plhs[0]);
    G       = mxGetPr(GArray);
    
    /* Do the actual computations in a subroutine */
    LengthConstraints(C,G,Coords,NVerts,Edges,NEdges,Lengths,mxGetScalar(prhs[3]));
    
    transposeMatrix(mxGetPr(plhs[1]),G,n3,NEdges);
    
    mxDestroyArray(GArray);         
    return;
    
}
