% Copyright (C) <2012>  <Aydin Varol, Mathieu Salzmann, Pascal Fua and Raquel Urtasun>
% 
% Implementation of CLVM method described in:
% A. Varol, M. Salzmann, P. Fua and R. Urtasun, A Constrained Latent Variable Model, 
% Conference on Computer Vision and Pattern Recognition, June 2012.
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the version 3 of the GNU General Public License
% as published by the Free Software Foundation.
% 
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
% General Public License for more details.       
% You should have received a copy of the GNU General Public License
% along with this program. If not, see <http://www.gnu.org/licenses/>.
%
% Aydin Varol, CVLab-EPFL, 2012.
% aydin.varol@gmail.com, www.aydinvarol.com


clc
clear all
close all

%% set seed
% initialization
rand('seed',123456);
randn('seed',123456);

%% config
   
% dataset
gamma= 1e-3;
alpha = 1e-3;
dataset = 'mesh9x9';

 
resPath = 'res.mat';
saveResultsP = 0;

% number of examples
numTests = 10;
numData =  20;
numValid = 20;
numTrial = 1;
sampleValidationDataP = 0;
sampleTestDataP = 0;

% kern type
kernType = 'rbf'; % poly or rbf

% strategy
trainSum = 0; % train with all the constraints
runMVCP = 0; % most violated constraints selection

% type of constraints
trainEq = 1;
trainIneq = 0;
squaredConstP = 0;

% optimization settings
vert2vertErrorP = 1;
displayRunDetailsP = 0;
pickBestIterP = 1;
numIt = 10;

% for MVC experiment
if(runMVCP)
    Vstart = numValid(1);
    if(length(numValid)>2)
        Vincreament = numValid(2)-numValid(1);
    else
        Vincreament = Vstart;
    end
end

if(saveResultsP)
   resFile = resPath;  
end


%% load data (LVM)

[Xall Yall Y Y0 Mesh My Sy dataIndsAll validIndsAll testInds M Nx Ny PCs Var] = loadDataLVM(dataset,numData,numValid,numTests,numTrial,sampleValidationDataP);


%% make constraints

if(strcmp(dataset(1:4),'mesh') )    
    if(trainSum==0)
        constrFunHandle = @(Y)LengthConstraintsMex(Y,(Mesh.Edges)',Mesh.Lengths,squaredConstP);
        constrFunHandleActualLength = @(Y)LengthConstraintsMex(Y,(Mesh.Edges)',Mesh.Lengths,0);
    else
        edges2sum = makeSumConstraintsMesh(Nx,Ny,Mesh.Edges);
        singleConstrFunHandle = @(Y)LengthConstraintsMex(Y,(Mesh.Edges)',Mesh.Lengths,squaredConstP);
        singleConstrFunHandleActualLength = @(Y)LengthConstraintsMex(Y,(Mesh.Edges)',Mesh.Lengths,0);
        constrFunHandle = @(Y) sumConstraints(Y,edges2sum,singleConstrFunHandle);
        constrFunHandleActualLength = @(Y)sumConstraints(Y,edges2sum,singleConstrFunHandleActualLength);
    end    
else
    error('unknown dataset');
end


%% global output definitions

REGP = zeros(length(numData),length(numValid));
REPCA = zeros(length(numData),length(numValid));
RECEq = zeros(length(numData),length(numValid));
RECIneq = zeros(length(numData),length(numValid));

CEGPEq = zeros(length(numData),length(numValid));
CEGPIneq = zeros(length(numData),length(numValid));
CEPCAEq = zeros(length(numData),length(numValid));
CEPCAIneq = zeros(length(numData),length(numValid));
CECEq = zeros(length(numData),length(numValid));
CECIneq = zeros(length(numData),length(numValid));

STDREGP = zeros(length(numData),length(numValid));
STDREPCA = zeros(length(numData),length(numValid));
STDRECEq = zeros(length(numData),length(numValid));
STDRECIneq = zeros(length(numData),length(numValid));

STDCEGPEq = zeros(length(numData),length(numValid));
STDCEGPIneq = zeros(length(numData),length(numValid));
STDCEPCAEq = zeros(length(numData),length(numValid));
STDCEPCAIneq = zeros(length(numData),length(numValid));
STDCECEq = zeros(length(numData),length(numValid));
STDCECIneq = zeros(length(numData),length(numValid));

% models
allModelsGP = cell(length(numData),length(numValid),numTrial);
allModelsCEq = cell(length(numData),length(numValid),numTrial);
allModelsCIneq = cell(length(numData),length(numValid),numTrial);


%% training constraint and unconstraint (GP) cases

for nd = 1 : length(numData)
    
    for nv = 1 : length(numValid)
        
        % local output definitions
        if(1)
            localREGP = zeros(numTrial,1);
            localREPCA = zeros(numTrial,1);
            localRECEq = zeros(numTrial,1);
            localRECIneq = zeros(numTrial,1);
            
            localCEGPEq = zeros(numTrial,1);
            localCEGPIneq = zeros(numTrial,1);
            localCEPCAEq = zeros(numTrial,1);
            localCEPCAIneq = zeros(numTrial,1);
            localCECEq = zeros(numTrial,1);
            localCECIneq = zeros(numTrial,1);
        end
        
        
        
        % trials
        for ti = 1 : numTrial
            
            fprintf('\rN=%d, V=%d, trial=%d\n',numData(nd),numValid(nv),ti);
            
            dataInds = dataIndsAll{nd}(ti,:);
            
            Xlabeled = Xall(dataInds,:);
            Ylabeled = Y(dataInds,:);
            
            if(sampleValidationDataP==1)
                error('not implemented yet');
            else
                validInds = validIndsAll{nv}(ti,:);
                Xunlabeled = Xall(validInds,:);
            end
                        
            % test data
            if(sampleTestDataP)
                error('not implemented yet');
            else
                Xtest = Xall(testInds,:);
                Ytest = Yall(testInds,:);
            end
            
            
            % set-up the kernel
            if(strcmp(kernType,'rbf'))
                GPKern = {'rbf','white'};
                opts = cell(2,1);
                %invWidth = 1/mean(mean(dist2(Xlabeled,Xlabeled)));
                invWidth = 1/mean(mean(dist2([Xlabeled;Xunlabeled],[Xlabeled;Xunlabeled])));
                opts{1} = 1/mean(mean(dist2(Xlabeled,Xlabeled)));
                opts{2} = gamma;
                myKern = @(X1,X2) computeRbfKern(X1,X2,invWidth);
            elseif(strcmp(kernType,'poly'))
                Xlabeled = applyQuadPhi(Xlabeled);
                Xunlabeled = applyQuadPhi(Xunlabeled);
                Xtest = applyQuadPhi(Xtest);
                GPKern = {'lin','white'};
                myKern = @computeLinKern;
                opts = cell(2,1);
                opts{1} = 1;
                opts{2} = gamma;
            else
                error('unknown kernel type');
            end
            
            
            %%%%%%%%%%%%%%%%%%% TRAININGS & PREDICTIONS %%%%%%%%%%%%%%%%%%
            
            % 1. training unconstrained GP
            modelGP = learnGP(Xlabeled,Ylabeled,GPKern,opts);
            predGP = predictGP(modelGP,Xtest,My',diag(Sy)') ;
            YGP = predGP+repmat(Y0,size(Xtest,1),1);
            
            % 2. training constrained model
            predGPhat = predictGP(modelGP,Xunlabeled,My',diag(Sy)');
            Yhat = predGPhat+repmat(Y0,size(Xunlabeled,1),1);
            if(trainEq)
                if(runMVCP==0)
                    modelCEq = learnConstraintKernelMex(Xlabeled,Ylabeled,Xunlabeled,myKern,constrFunHandle,Yhat,Y0,gamma,numIt,pickBestIterP,displayRunDetailsP);
                    if(strcmp(kernType,'rbf'))
                        modelCEq.invWidth = invWidth;
                    end
                else
                    trainingFunHandle = @(Xlabeled,Xunlabeled,Yhat,numIt)learnConstraintKernelMex(Xlabeled,Ylabeled,Xunlabeled,myKern,constrFunHandle,Yhat,Y0,gamma,numIt,pickBestIterP,displayRunDetailsP);
                    [allModelsMVC modelCEq] = learnConstraintKernelMVCMex(Xlabeled,Xunlabeled,Yhat,numIt,myKern,My,Sy,Y0,constrFunHandle,trainingFunHandle,Vstart,Vincreament);
                    if(strcmp(kernType,'rbf'))
                        modelCEq.invWidth = invWidth;
                    end
                end
                predCEq = predictConstraintKernel(Xtest,modelCEq,myKern,My,Sy);
                YCEq = predCEq + repmat(Y0,size(Xtest,1),1);
            end
            if(trainIneq)
                if(runMVCP==0)
                    modelCIneq = learnConstraintKernelIneqMex(Xlabeled,Ylabeled,Xunlabeled,myKern,constrFunHandle,Yhat,Y0,gamma,alpha,numIt,pickBestIterP,displayRunDetailsP);
                    if(strcmp(kernType,'rbf'))
                        modelCIneq.invWidth = invWidth;
                    end
                else
                    trainingFunHandle = @(Xlabeled,Xunlabeled,Yhat,numIt)learnConstraintKernelIneqMex(Xlabeled,Ylabeled,Xunlabeled,myKern,constrFunHandle,Yhat,Y0,gamma,alpha,numIt,pickBestIterP,displayRunDetailsP);
                    [allModelsMVC modelCIneq] = learnConstraintKernelMVCMex(Xlabeled,Xunlabeled,Yhat,numIt,myKern,My,Sy,Y0,constrFunHandle,trainingFunHandle,Vstart,Vincreament);
                    if(strcmp(kernType,'rbf'))
                        modelCIneq.invWidth = invWidth;
                    end
                end
                predCIneq = predictConstraintKernel(Xtest,modelCIneq,myKern,My,Sy);
                YCIneq = predCIneq + repmat(Y0,size(Xtest,1),1);
            end
            
            % 3. PCA
            pcaBases = computePCABases(Ylabeled);
            predPCA = reconstructFromBasis(Xtest,pcaBases(:,1:M),Y0);
            YPCA = predPCA+repmat(Y0,size(Xtest,1),1);
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            
            %%%%%%%%%%%%%%%%% ERRORS %%%%%%%%%%%%%%%%%%%%%%
            
            % 1. errors unconstrained GP
            if(sampleTestDataP==0)
                localREGP(ti) =  mean(computeReconstructionError(YGP,Ytest,vert2vertErrorP));
            end
            if(trainEq)
                localCEGPEq(ti) = mean(computeConstError(YGP,constrFunHandleActualLength));
            end
            if(trainIneq)
                localCEGPIneq(ti) = mean(computeIneqConstError(YGP,constrFunHandleActualLength));
            end
            
            % 2. errors on constrained model
            if(trainEq)
                if(sampleTestDataP==0)
                    localRECEq(ti) =  mean(computeReconstructionError(YCEq,Ytest,vert2vertErrorP));
                end
                localCECEq(ti) = mean(computeConstError(YCEq,constrFunHandleActualLength));
            end
            if(trainIneq)
                if(sampleTestDataP==0)
                    localRECIneq(ti) =  mean(computeReconstructionError(YCIneq,Ytest,vert2vertErrorP));
                end
                localCECIneq(ti) = mean(computeIneqConstError(YCIneq,constrFunHandleActualLength));
            end
              
            % 3. errors PCA
            if(sampleTestDataP==0)
                localREPCA(ti) =  mean(computeReconstructionError(YPCA,Ytest,vert2vertErrorP));
            end
            if(trainEq)
                localCEPCAEq(ti) = mean(computeConstError(YPCA,constrFunHandleActualLength));
            end
            if(trainIneq)
                localCEPCAIneq(ti) = mean(computeIneqConstError(YPCA,constrFunHandleActualLength));
            end
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%% LOCAL SAVES %%%%%%%%%%%%%%%%%%%%%%%%
            allModelsGP{nd,nv,ti} = modelGP;
            if(trainEq)
                allModelsCEq{nd,nv,ti} = modelCEq;
            end
            if(trainIneq)
                allModelsCIneq{nd,nv,ti} = modelCIneq;
            end
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            % %%%%%%%%%%%%%%%% FREE MEMORY %%%%%%%%%%%%%%%%%%
            clear modelGP;
            clear modelCEq;
            clear modelCIneq;
            clear allModelsMVC;
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            clc;
            
        end %end trials
                
        % average errors over trials
        REGP(nd,nv) = mean(localREGP);
        REPCA(nd,nv) = mean(localREPCA);
        RECEq(nd,nv) = mean(localRECEq);
        RECIneq(nd,nv) = mean(localRECIneq);
        
        CEGPEq(nd,nv) = mean(localCEGPEq);
        CEGPIneq(nd,nv) = mean(localCEGPIneq);
        CEPCAEq(nd,nv) = mean(localCEPCAEq);
        CEPCAIneq(nd,nv) = mean(localCEPCAIneq);
        CECEq(nd,nv) = mean(localCECEq);
        CECIneq(nd,nv) = mean(localCECIneq);
        
        STDREGP(nd,nv) = std(localREGP);
        STDREPCA(nd,nv) = std(localREPCA);
        STDRECEq(nd,nv) = std(localRECEq);
        STDRECIneq(nd,nv) = std(localRECIneq);
        
        STDCEGPEq(nd,nv) = std(localCEGPEq);
        STDCEGPIneq(nd,nv) = std(localCEGPIneq);
        STDCEPCAEq(nd,nv) = std(localCEPCAEq);
        STDCEPCAIneq(nd,nv) = std(localCEPCAIneq);
        STDCECEq(nd,nv) = std(localCECEq);
        STDCECIneq(nd,nv) = std(localCECIneq);                
        
    end
end


%% displays

REGP
REPCA
if(trainEq)
    RECEq
end
if(trainIneq)
    RECIneq
end

if(trainEq)
    CEGPEq
    CEPCAEq
    CECEq
end
if(trainIneq)
    CEGPIneq
    CEPCAIneq
    CECIneq    
end


%% global saves
if(saveResultsP)
    save(resFile,'allModelsGP','Xall','Yall','Y','Y0','Mesh','My','Sy','dataIndsAll','validIndsAll','testInds',...
        'alpha','gamma','dataset','resPath','sampleValidationDataP','sampleTestDataP','kernType',...
        'trainSum','runMVCP','trainEq','trainIneq','squaredConstP','vert2vertErrorP','pickBestIterP','numIt',...
        'M','Nx','Ny','PCs','Var',...
        '-regexp', '^RE|^CE|^STDRE|^STDCE',...
        '-regexp', '^allModels');
end

